/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.util;

import dk.sds.nsp.security.SecurityContext;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Component
public class ProxyDataProvider {
    public static final String AUTH_LEVEL_SYSTEM = "3";
    public static final String AUTH_LEVEL_USER = "4";

    private SecurityContextProvider securityContextProvider;

    public ProxyDataProvider(@Autowired SecurityContextProvider securityContextProvider) {
        this.securityContextProvider = securityContextProvider;
    }

    public ProxyData getProxyData() {
        SecurityContext securityContext = securityContextProvider.getSecurityContext();
        validateSecurityContext(securityContext);
        return mapSecurityContextToProxyData(securityContext);
    }

    private ProxyData mapSecurityContextToProxyData(SecurityContext securityContext) {
        ProxyData proxyData = new ProxyData(getUserType(securityContext));

        if (securityContext.getMessage().isPresent()) {
            mapDataFromMessage(proxyData, securityContext.getMessage().get());
        }

        // A non-present ActingUser corresponds to authentication level 3 (system) ...
        if (securityContext.getActingUser().isPresent()) {
            mapDataFromUser(proxyData, securityContext.getActingUser().get());
            proxyData.setAuthenticationLevel(AUTH_LEVEL_USER);
        }
        else {
            proxyData.setAuthenticationLevel(AUTH_LEVEL_SYSTEM);
        }

        if (securityContext.getClient().isPresent()) {
            mapDataFromClient(proxyData, securityContext.getClient().get());
        }

        if (securityContext.getOrganisation().isPresent()) {
            mapDataFromOrganisation(proxyData, securityContext.getOrganisation().get());
        }

        return proxyData;
    }

    private void mapDataFromMessage(ProxyData proxyData, SecurityContext.Message message) {
        proxyData.setMessageId(message.getIdentifier().orElse(null));
        proxyData.setFlowId(message.getConversationIdentifier().orElse(null));
    }

    private void mapDataFromUser(ProxyData proxyData, SecurityContext.User user) {
        if (user.getIdentifierFormat() == SecurityContext.User.PersonIdentifierFormat.CPR) {
            proxyData.setCprnr(user.getIdentifier());
        }
        proxyData.setGivenName(user.getGivenName().orElse(null));
        proxyData.setSurName(user.getSurname().orElse(null));
        if (user.getCredentials().isPresent()) {
            SecurityContext.User.Credentials credentials = user.getCredentials().get();
            proxyData.setAuthorizationCode(credentials.getAuthorizationCode().orElse(null));
            proxyData.setEducationCode(credentials.getEducationCode().orElse(null));
        }
    }

    private void mapDataFromClient(ProxyData proxyData, SecurityContext.Client client) {
        proxyData.setItSystemName(client.getName().orElse(null));
    }

    private void mapDataFromOrganisation(ProxyData proxyData, SecurityContext.Organisation organisation) {
        proxyData.setCareProviderID(organisation.getIdentifier());
        proxyData.setCareProviderName(organisation.getName().orElse(null));
        if(organisation.getIdentifierFormat() != null) {
            switch(organisation.getIdentifierFormat()) {
                case CVR:
                    proxyData.setCareProviderType("medcom:cvrnumber");
                    break;
                default:
                    throw new IllegalArgumentException(String.format("Unknown IdentifierFormat: %s", organisation.getIdentifierFormat().toString()));
            }
        }
    }

    private ProxyData.UserType getUserType(SecurityContext securityContext) {
        if (!securityContext.getActingUser().isPresent() || securityContext.getActingUser().get().getUserType() == null) {
            return ProxyData.UserType.SystemUser;
        }
        switch (securityContext.getActingUser().get().getUserType()) {
            case HealthcareProfessional:
                return ProxyData.UserType.HealthCareProfessional;
            case Citizen:
                return ProxyData.UserType.Citizen;
            default:
                return ProxyData.UserType.SystemUser;
        }
    }

    private void validateSecurityContext(SecurityContext securityContext) {
        if (securityContext == null || !securityContext.getTicket().isPresent() || !securityContext.getTicket().get().isValid()) {
            FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
        }
    }
}
