/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.util;

import dk.sds.nsp.security.SecurityContext;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.junit.MockitoJUnitRunner;

import java.util.Optional;

import static org.junit.Assert.*;

@RunWith(MockitoJUnitRunner.class)
public class ProxyDataProviderTest {
    @InjectMocks
    private ProxyDataProvider subject;

    @Mock
    private SecurityContextProvider securityContextProvider;

    @Test
    public void getProxyData_flowId() {
        // Arrange
        setupSecurityContext(buildMessage("foo", "bar"), null, null);

        // Act
        ProxyData result = subject.getProxyData();

        // Assert
        assertEquals("foo", result.getFlowId());
    }

    @Test
    public void getProxyData_messageId() {
        // Arrange
        setupSecurityContext(buildMessage("foo", "bar"), null, null);

        // Act
        ProxyData result = subject.getProxyData();

        // Assert
        assertEquals("bar", result.getMessageId());
    }

    @Test
    public void getProxyData_authenticationLevel3() {
        // Arrange
        setupSecurityContext(buildMessage("foo", "bar"), null, null);

        // Act
        ProxyData result = subject.getProxyData();

        // Assert
        assertEquals("3", result.getAuthenticationLevel());
    }

    @Test
    public void getProxyData_authenticationLevel4() {
        // Arrange
        setupSecurityContext(buildMessage("foo", "bar"), null, buildActingUser("0101010101"));

        // Act
        ProxyData result = subject.getProxyData();

        // Assert
        assertEquals("4", result.getAuthenticationLevel());
    }

    @Test
    public void getProxyData_cprnr() {
        // Arrange
        setupSecurityContext(buildMessage("foo", "bar"), null, buildActingUser("0101010101"));

        // Act
        ProxyData result = subject.getProxyData();

        // Assert
        assertEquals("0101010101", result.getCprnr());
    }

    @Test
    public void getProxyData_careProviderName() {
        // Arrange
        setupSecurityContext(null, buildOrganisation("Lægehuset", "12345678", SecurityContext.Organisation.OrganisationIdentifierFormat.CVR), null);

        // Act
        ProxyData result = subject.getProxyData();

        // Assert
        assertEquals("Lægehuset", result.getCareProviderName());
    }

    @Test
    public void getProxyData_careProviderId() {
        // Arrange
        setupSecurityContext(null, buildOrganisation("Lægehuset", "12345678", SecurityContext.Organisation.OrganisationIdentifierFormat.CVR), null);

        // Act
        ProxyData result = subject.getProxyData();

        // Assert
        assertEquals("12345678", result.getCareProviderID());
    }

    @Test
    public void getProxyData_careProviderType() {
        // Arrange
        setupSecurityContext(null, buildOrganisation("Lægehuset", "12345678", SecurityContext.Organisation.OrganisationIdentifierFormat.CVR), null);

        // Act
        ProxyData result = subject.getProxyData();

        // Assert
        assertEquals("medcom:cvrnumber", result.getCareProviderType());
    }

    private void setupSecurityContext(SecurityContext.Message message, SecurityContext.Organisation organisation, SecurityContext.User actingUser) {
        SecurityContext context = Mockito.mock(SecurityContext.class);

        if(message != null) {
            Mockito.when(context.getMessage()).thenReturn(Optional.of(message));
        }
        if(organisation != null) {
            Mockito.when(context.getOrganisation()).thenReturn(Optional.of(organisation));
        }
        if(actingUser != null) {
            Mockito.when(context.getActingUser()).thenReturn(Optional.of(actingUser));
        }

        SecurityContext.Ticket ticket = Mockito.mock(SecurityContext.Ticket.class);
        Mockito.when(ticket.isValid()).thenReturn(true);
        Mockito.when(context.getTicket()).thenReturn(Optional.of(ticket));

        Mockito.when(securityContextProvider.getSecurityContext()).thenReturn(context);
    }

    private SecurityContext.Message buildMessage(String flowId, String messageId) {
        SecurityContext.Message message = Mockito.mock(SecurityContext.Message.class);

        Mockito.when(message.getConversationIdentifier()).thenReturn(Optional.of(flowId));
        Mockito.when(message.getIdentifier()).thenReturn(Optional.of(messageId));

        return message;
    }

    private SecurityContext.Organisation buildOrganisation(String name, String identifier, SecurityContext.Organisation.OrganisationIdentifierFormat identifierFormat) {
        SecurityContext.Organisation organisation = Mockito.mock(SecurityContext.Organisation.class);

        Mockito.when(organisation.getName()).thenReturn(Optional.of(name));
        Mockito.when(organisation.getIdentifier()).thenReturn(identifier);
        Mockito.when(organisation.getIdentifierFormat()).thenReturn(identifierFormat);

        return organisation;
    }

    private SecurityContext.User buildActingUser(String cprnr) {
        SecurityContext.User actingUser = Mockito.mock(SecurityContext.User.class);

        Mockito.when(actingUser.getIdentifierFormat()).thenReturn(SecurityContext.User.PersonIdentifierFormat.CPR);
        Mockito.when(actingUser.getIdentifier()).thenReturn(cprnr);

        return actingUser;
    }
}