/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort;

import dk.sundhedsdatastyrelsen.fsk._2020._07._28.DGWSFault;
import dk.sundhedsdatastyrelsen.fsk._2020._07._28.PersonalDataCardPortType;
import dk.stamkort.dgws.SosiIdCardHandler;
import dk.stamkort.dgws.UserIdCardHelper;
import dk.stamkort.util.configuration.Configuration;
import dk.stamkort.util.skr.SkrFacade;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.util.EntityUtils;
import org.hl7.v3.*;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;

import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.transform.stream.StreamSource;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;

import static org.junit.Assert.*;

public class OnDemandServiceITCase {
    Configuration configurationProvider = new Configuration();
    private String skrEndpoint = "http://test1-cnsp.ekstern-test.nspop.dk:8080/skr/skr";
    SkrFacade skrFacade;

    private static final String ENCODING = "UTF-8";
    public static final String TEST_PERSON = "0509900267";

    private String host = "localhost";
    private String port = "8080";
    private String docid = "urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000003";

    private SosiIdCardHandler sosiIdCardHandler = UserIdCardHelper.createUserIdCardHelper();

    public OnDemandServiceITCase() {
        String host = System.getProperty("FSK_TEST_HOST");
        String port = System.getProperty("FSK_TEST_PORT");
        String docid = System.getProperty("FSK_TEST_DOCID");
        String skrEndpoint = System.getProperty("SKR_ENDPOINT");

        if (host != null && host.length() >0 ) {
            this.host = host;
        }

        if (port != null && port.length() > 0) {
            this.port = port;
        }

        if (docid != null && docid.length() > 0) {
            this.docid = docid;
        }

        if (skrEndpoint != null && skrEndpoint.length() > 0) {
            this.skrEndpoint = skrEndpoint;
        }

        this.skrFacade = new SkrFacade(configurationProvider.createSkrClient(this.skrEndpoint));
    }

    @Before
    public void setup() throws Exception {
        cleanup();
        setupTestData();
    }

    @Test
    public void callOnDemandService_PDC_1_0ClinicalDocument_() throws Exception {
        InputStream stream = getClass().getResourceAsStream("/testrequest.xml");
        SOAPMessage soapMessage = MessageFactory.newInstance().createMessage(null, stream);

        System.out.println("Retrieving DGWS headers.");
        try {
            Document sosi = sosiIdCardHandler.getSosiDocument();
            NodeList children = sosi.getDocumentElement().getFirstChild().getChildNodes();
            SOAPHeader soapHeader = soapMessage.getSOAPHeader();

            for(int i = 0; i < children.getLength(); i++) {
                Node element = children.item(i);
                soapHeader.appendChild(soapHeader.getOwnerDocument().importNode(element, true));
            }
        } catch (Throwable ex) {
            System.err.println("Retrieving DGWS headers failed.");
            ex.printStackTrace();
        }

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        soapMessage.writeTo(baos);
        String xml = baos.toString(ENCODING);

        if (docid != null) {
            xml = xml.replace("urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000000", docid);
        }

        System.out.println("Request: " + xml);
        String response = httpWrite("urn:ihe:iti:2007:RetrieveDocumentSet", xml);
        System.out.println("Response: " + response);

        assertNotNull(response);
        assertTrue(response.contains("ClinicalDocument"));
    }

    @Test
    public void callOnDemandService_PDC_2_0ClinicalDocument_correctIds() throws Exception {
        InputStream stream = getClass().getResourceAsStream("/testrequest2.xml");
        SOAPMessage soapMessage = MessageFactory.newInstance().createMessage(null, stream);

        System.out.println("Retrieving DGWS headers.");
        try {
            Document sosi = sosiIdCardHandler.getSosiDocument();
            NodeList children = sosi.getDocumentElement().getFirstChild().getChildNodes();
            SOAPHeader soapHeader = soapMessage.getSOAPHeader();

            for(int i = 0; i < children.getLength(); i++) {
                Node element = children.item(i);
                soapHeader.appendChild(soapHeader.getOwnerDocument().importNode(element, true));
            }
        } catch (Throwable ex) {
            System.err.println("Retrieving DGWS headers failed.");
            ex.printStackTrace();
        }

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        soapMessage.writeTo(baos);
        String xml = baos.toString(ENCODING);

        if (docid != null) {
            xml = xml.replace("urn:sds:fsk:stamkort:00000000-0000-0000-0000-000000000001", docid);
        }

        System.out.println("Request: " + xml);
        String responseA = httpWrite("urn:ihe:iti:2007:RetrieveDocumentSet", xml);
        System.out.println("Response: " + responseA);

        String xmlResponseA = responseA.substring(responseA.indexOf("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?><ClinicalDocument moodCode=\"EVN\""), responseA.indexOf("</ClinicalDocument>")+"</ClinicalDocument>".length());
        assertNotNull(responseA);
        assertTrue(responseA.contains("ClinicalDocument"));

        JAXBContext contextA = JAXBContext.newInstance(POCDMT000040ClinicalDocument.class);
        JAXBElement<POCDMT000040ClinicalDocument> rA = contextA.createUnmarshaller().unmarshal(new StreamSource(new StringReader(xmlResponseA)), POCDMT000040ClinicalDocument.class);

        String languageIdA = rA.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(5).getObservation().getIds().get(0).getExtension();
        String tempAddrIdA = rA.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(6).getObservation().getIds().get(0).getExtension();
        String dentistIdA = rA.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(7).getObservation().getIds().get(0).getExtension();

        // Run again
        String responseB = httpWrite("urn:ihe:iti:2007:RetrieveDocumentSet", xml);

        String xmlResponseB = responseB.substring(responseB.indexOf("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?><ClinicalDocument moodCode=\"EVN\""), responseB.indexOf("</ClinicalDocument>")+"</ClinicalDocument>".length());
        assertNotNull(responseB);
        assertTrue(responseB.contains("ClinicalDocument"));

        JAXBContext contextB = JAXBContext.newInstance(POCDMT000040ClinicalDocument.class);
        JAXBElement<POCDMT000040ClinicalDocument> rB = contextB.createUnmarshaller().unmarshal(new StreamSource(new StringReader(xmlResponseB)), POCDMT000040ClinicalDocument.class);

        String languageIdB = rB.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(5).getObservation().getIds().get(0).getExtension();
        String tempAddrIdB = rB.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(6).getObservation().getIds().get(0).getExtension();
        String dentistIdB = rB.getValue().getComponent().getStructuredBody().getComponents().get(0).getSection().getEntries().get(7).getObservation().getIds().get(0).getExtension();

        // Assert the same Ids are provided in both calls
        assertTrue(tempAddrIdA.equals(tempAddrIdB));
        assertTrue(dentistIdA.equals(dentistIdB));
        assertTrue(languageIdA.equals(languageIdB));
    }

    private String httpWrite(String soapAction, String xmldata) throws Exception {
        HttpClient httpClient = new DefaultHttpClient();

        String endpoint = "http://" + host + ":" + port + "/fsk/services/fsk";
        System.out.println(" Calling " + endpoint + ", SOAPAction=" + soapAction);
        HttpPost request = new HttpPost(endpoint);
        request.addHeader("SOAPAction", "\"" + soapAction + "\"");

        StringEntity stringEntity = new StringEntity(xmldata, ENCODING);
        stringEntity.setContentType("text/xml");
        request.setEntity(stringEntity);

        HttpResponse response = httpClient.execute(request);
        return EntityUtils.toString(response.getEntity(), ENCODING);
    }

    private void setupTestData() throws DGWSFault {
        // Spoken language
        skrFacade.addLanguage(TEST_PERSON, "Georg", "Olesen", "de");
        // Temporary Address
        skrFacade.addTemporaryAddress(TEST_PERSON);
        // Dentist
       skrFacade.addDentist(TEST_PERSON);
    }

    private void cleanup() throws Exception {
        // Find id'er fra skr
        POCDMT000040ClinicalDocument clinicalDocument = skrFacade.getSkr(TEST_PERSON);

        // sprog fra TEST_PERSON
        List<String> languageIds = getLanguageIds(clinicalDocument);
        for (String languageId : languageIds) {
            skrFacade.deleteLanguage(TEST_PERSON, languageId);
        }
        // midlertidig adresse fra TEST_PERSON
        List<String> tempAddrIds = getTempAddrIds(clinicalDocument);
        for (String tempAddrId: tempAddrIds) {
            skrFacade.deleteTemporaryAddress(TEST_PERSON, tempAddrId);
        }
        // tandlæge fra TEST_PERSON
        List<String> dentistIds = getDentistsIds(clinicalDocument);
        for (String dentistId: dentistIds) {
            skrFacade.deleteDentist(TEST_PERSON, dentistId);
        }
    }

    private List<String> getDentistsIds(POCDMT000040ClinicalDocument clinicalDocument) {
        List<String> dentistIds = new ArrayList<>();
        if (clinicalDocument == null) {
            return dentistIds;
        }

        if (clinicalDocument.getComponent().getStructuredBody() == null) {
            return dentistIds;
        }

        for (POCDMT000040Component3 component : clinicalDocument.getComponent().getStructuredBody().getComponents()) {
            for (POCDMT000040Entry entri : component.getSection().getEntries()) {
                if (entri.getHealthProvider() != null) {
                    dentistIds.add(entri.getHealthProvider().getId().getExtension());
                }
            }
        }
        return dentistIds;
    }

    private List<String> getLanguageIds(POCDMT000040ClinicalDocument clinicalDocument) {
        List<String> languageIds = new ArrayList<>();
        if (clinicalDocument == null) {
            return languageIds;
        }

        if (clinicalDocument.getComponent().getStructuredBody() == null) {
            return languageIds;
        }

        for (POCDMT000040Component3 component : clinicalDocument.getComponent().getStructuredBody().getComponents()) {
            for (POCDMT000040Entry entri : component.getSection().getEntries()) {
                if (entri.getLanguage() != null) {
                    languageIds.add(entri.getLanguage().getId().getExtension());
                }
            }
        }
        return languageIds;
    }

    private List<String> getTempAddrIds(POCDMT000040ClinicalDocument clinicalDocument) {
        List<String> tempAddrIds = new ArrayList<>();
        if (clinicalDocument == null) {
            return tempAddrIds;
        }

        if (clinicalDocument.getComponent().getStructuredBody() == null) {
            return tempAddrIds;
        }

        for (POCDMT000040Component3 component : clinicalDocument.getComponent().getStructuredBody().getComponents()) {
            for (POCDMT000040Entry entri : component.getSection().getEntries()) {
                if (entri.getTemporaryAddress() != null) {
                    tempAddrIds.add(entri.getTemporaryAddress().getId().getExtension());
                }
            }
        }
        return tempAddrIds;
    }
}
