/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
/**
* The MIT License
*   
* Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
*   
* Copyright (C) 2018 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
*   
*   
* Permission is hereby granted, free of charge, to any person obtaining a copy of
* this software and associated documentation files (the "Software"), to deal in
* the Software without restriction, including without limitation the rights to
* use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*   
*   
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*   
*   
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*/ 
package dk.stamkort.util.dgws;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.xml.XmlUtil;
import org.springframework.core.io.Resource;
import org.w3c.dom.Document;

import java.io.IOException;
import java.io.InputStream;
import java.security.*;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Calendar;
import java.util.Date;
import java.util.Properties;

public abstract class DgwsNspUserContext {
	protected UserInfo userInfo;

	protected CredentialVaultWrapper credentialVault;
	
	protected STSRequestHelper requestHelper;
	
	protected SOSIFactory sosiFactory;

	protected String userOrganizationCode;
	
	protected String responsibleUserCivilRegistrationNumber;
	
	protected IDCard cachedCard;
	
	public DgwsNspUserContext(STSRequestHelper requestaHelper, Resource keystore, String password, String alias, String patientId, UserInfo userInfo, String userOrganizationCode, String responsibleUserCivilRegistrationNumber, boolean consentOverride) throws IOException {
		this.x509 = getX509Wrapper(keystore, password, alias);
		this.patientId = patientId;
		this.userInfo = userInfo;
		this.responsibleUserCivilRegistrationNumber = responsibleUserCivilRegistrationNumber;
		this.userOrganizationCode = userOrganizationCode;
		this.credentialVault = new CredentialVaultWrapper(x509.getCertificate(), x509.getPrivateKey());

		Properties props = new Properties(System.getProperties());
		props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, Boolean.toString(true));

		this.requestHelper = requestaHelper;
		this.sosiFactory = new SOSIFactory(new SOSITestFederation(new Properties()), this.credentialVault, props);
	}

	public Document getSosiIdCardRequest() {
		
		SecurityTokenRequest securityTokenRequest = sosiFactory.createNewSecurityTokenRequest();
		securityTokenRequest.setIDCard(createNewIdCard(sosiFactory));
		Document doc = securityTokenRequest.serialize2DOMDocument();
		
		if (signSosiCardRequests()) {
			SignatureConfiguration signatureConfiguration = new SignatureConfiguration(new String[] { IDValues.IDCARD }, IDValues.IDCARD, IDValues.id);
			SignatureUtil.sign(SignatureProviderFactory.fromCredentialVault(credentialVault), doc, signatureConfiguration);
		}
		
		return doc;		
	}

	protected abstract boolean signSosiCardRequests();

	public abstract SystemIDCard createNewIdCard(SOSIFactory sosi);

	public IDCard getToken() throws IOException {

		Calendar cal = Calendar.getInstance();
//		cal.add(Calendar.MINUTE, 5); 
		cal.add(Calendar.HOUR_OF_DAY, 23); //needs to be syncronized with the idcard.max.age.hours in application.properties. If age.hours set to 1, then hour_of_day needs to be 23. It will then run for aprox 50 min on same idcard.
		Date timeFromNow = cal.getTime();

		if (cachedCard != null && cachedCard.getExpiryDate().after(timeFromNow)) {
			return cachedCard;
		}
		
		Document requestXml = getSosiIdCardRequest();
		String requestXmlString = XmlUtil.node2String(requestXml, false, true);
		
		String responseXml = requestHelper.sendRequest(sosiFactory, requestXmlString);
		SecurityTokenResponse securityTokenResponse = sosiFactory.deserializeSecurityTokenResponse(responseXml);

		if (securityTokenResponse.isFault() || securityTokenResponse.getIDCard() == null) {
			throw new RuntimeException("No ID card :-(");
		}
		else {
			SystemIDCard stsSignedIdCard = (SystemIDCard) securityTokenResponse.getIDCard();
			cachedCard = stsSignedIdCard;
			return cachedCard;
		}
	}
	
	public SOSIFactory getSosiFactory() {
		return this.sosiFactory;
	}

	protected X509Wrapper x509;
	
	protected String patientId;
	
	public static X509Wrapper getX509Wrapper(Resource keystoreResource, String password, String alias) throws IOException {
		try {
			InputStream in = keystoreResource.getInputStream();
			KeyStore keyStore = KeyStore.getInstance("JKS");
			keyStore.load(in, password.toCharArray());
			X509Certificate certificate = (X509Certificate) keyStore.getCertificate(alias);
			PrivateKey privateKey = (PrivateKey) keyStore.getKey(alias, password.toCharArray());
			return new X509Wrapper(certificate, privateKey);
		} catch (KeyStoreException | NoSuchAlgorithmException | CertificateException | UnrecoverableKeyException e) {
			throw new IOException("Error loading certificate and private key from " + keystoreResource, e);
		}
	}
	
	public String getPatientId() {
		return this.patientId;
	}

	protected UserInfo getUserInfo() {
		return userInfo;
	}
}
