/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.util.skr;

import dk.sundhedsdatastyrelsen.fsk._2020._07._28.ObjectFactory;
import dk.sundhedsdatastyrelsen.fsk._2020._07._28.*;
import dk.sundhedsdatastyrelsen.fsk.extensions.*;

import org.hl7.v3.*;

import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Arrays;

public class SkrFacade {
    public static final String TEMPORORARY_ADDRESS_ID = "a0026c52-bd06-4603-9d69-9bf1d00973a9";
    private final PersonalDataCardPortType skrClient;
    private final org.hl7.v3.ObjectFactory objectFactory = new org.hl7.v3.ObjectFactory();

    public SkrFacade(PersonalDataCardPortType skrClient) {
        this.skrClient = skrClient;
    }

    public void addLanguage(String cpr, String givenName, String familiyName, String lang) throws DGWSFault {
        ObjectFactory languageObjectFactory = new ObjectFactory();

        CreateLanguageRequest request = languageObjectFactory.createCreateLanguageRequest();

        II ii = new II();
        ii.setAssigningAuthorityName("CPR");
        ii.setExtension(cpr);
        ii.setRoot("1.2.208.176.1.2");
        request.setId(ii);

        Language language = new Language();
        LanguageCodeType languageCode = new LanguageCodeType();
        languageCode.setValue(lang);
        languageCode.setCodeSystem("1.2.208.184.100.1");
        language.setLanguageCode(languageCode);

        request.setLanguage(language);

        // Data enterer
        POCDMT000040Author dataEnterer = new POCDMT000040Author();
        TS ts = new TS();
        ts.setValue(ZonedDateTime.now().format(DateTimeFormatter.ofPattern("yyyyMMddHHmmssZ"))); //
        dataEnterer.setTime(ts);
        POCDMT000040AssignedAuthor assignedAuthor = new POCDMT000040AssignedAuthor();
        II entererId = new II();
        entererId.setExtension("0309691444");
        entererId.setAssigningAuthorityName("CPR");
        entererId.setRoot("1.2.208.176.1.2");
        assignedAuthor.getIds().add(entererId);
        POCDMT000040Person assignedPerson = new POCDMT000040Person();
        PN pn = new PN();

        EnGiven enGiven = new EnGiven();
        enGiven.getContent().add(givenName);
        pn.getContent().add(new org.hl7.v3.ObjectFactory().createENGiven(enGiven));

        EnFamily enFamily = new EnFamily();
        enFamily.getContent().add(familiyName);

        pn.getContent().add(new org.hl7.v3.ObjectFactory().createENFamily(enFamily));

        assignedPerson.getNames().add(pn);
        assignedAuthor.setAssignedPerson(assignedPerson);
        dataEnterer.setAssignedAuthor(assignedAuthor);
        request.setDataEnterer(dataEnterer);

        skrClient.createLanguage20200728(request);
    }

    public void deleteLanguage(String cpr, String languageIdString) throws DGWSFault {
        if (languageIdString == null){
            return;
        }
        DeleteLanguageRequest request = new DeleteLanguageRequest();
        request.setId(createIdElement(cpr));


        II languageId = createId(languageIdString);

        request.setLanguageId(languageId);
        request.setDataEnterer(createDateEntererElement());

        try {
            skrClient.deleteLanguage20200728(request);
        } catch (Exception e) {
            if (!e.getMessage().contains("did not match any current data")) {
                throw e;
            }
        }
    }

    public void addTemporaryAddress(String cpr) throws DGWSFault {
        CreateTemporaryAddressRequest request = new CreateTemporaryAddressRequest();
        request.setId(createIdElement(cpr));
        TemporaryAddress address = new TemporaryAddress();
        II temporaryAddressId = createTemporaryId(TEMPORORARY_ADDRESS_ID);

        AD ad = createAddress("Temporary Street Name 20");
        SXCMTS useablePeriod = new SXCMTS();
        useablePeriod.setValue("20191001");

        SXCMTS end  = new SXCMTS();
        end.setValue("20191010");


        ad.getContent().add(objectFactory.createADUseablePeriod(useablePeriod));
        ad.getContent().add(objectFactory.createADUseablePeriod(end));
        address.setAddr(ad);

        request.setTemporaryAddress(address);
        request.setDataEnterer(createDateEntererElement());

        skrClient.createTemporaryAddress20200728(request);
    }

    public void deleteTemporaryAddress(String cpr, String tempAddrId) throws DGWSFault {
        if (tempAddrId == null){
            return;
        }
        DeleteTemporaryAddressRequest request = new DeleteTemporaryAddressRequest();
        request.setId(createIdElement(cpr));


        II temporaryAddressId = createTemporaryId(tempAddrId);

        request.setTemporaryAddressId(temporaryAddressId);
        request.setDataEnterer(createDateEntererElement());

        try {
            skrClient.deleteTemporaryAddress20200728(request);
        } catch (Exception e) {
            if (!e.getMessage().contains("did not match any current data")) {
                throw e;
            }
        }
    }

    private II createTemporaryId(String tempororaryAddressId) {
        II temporaryAddressId = new II();
        temporaryAddressId.setExtension(tempororaryAddressId);
        temporaryAddressId.setRoot("1.2.208.184.15.2");
        temporaryAddressId.setAssigningAuthorityName("FSK");
        return temporaryAddressId;
    }

    private II createId(String idString) {
        II id = new II();
        id.setExtension(idString);
        id.setRoot("1.2.208.184.15.2");
        id.setAssigningAuthorityName("FSK");
        return id;
    }

    private AD createAddress(String streeName) {
        AD address = new AD();
        address.setIsNotOrdered(false);
        address.getUses().add("H");
        AdxpStreetAddressLine streetAddress = new AdxpStreetAddressLine();
        streetAddress.getContent().add(streeName);
        AdxpPostalCode postal = new AdxpPostalCode();
        postal.getContent().add("8000");
        AdxpCity city = new AdxpCity();
        city.getContent().add("Aarhus C");
        AdxpCountry country = new AdxpCountry();
        country.getContent().add("Denmark");
        address.getContent().add(objectFactory.createADStreetAddressLine(streetAddress));
        address.getContent().add(objectFactory.createADPostalCode(postal));
        address.getContent().add(objectFactory.createADCity(city));
        address.getContent().add(objectFactory.createADCountry(country));

        return address;
    }

    public void addContactInformation(String cpr) throws DGWSFault {
        UpdateContactInformationRequest request = new UpdateContactInformationRequest();
        request.setId(createIdElement(cpr));
        request.getTelecoms().add(createTel(ContactType.H, "118"));
        request.getTelecoms().add(createTel(ContactType.WP, "911"));
        request.getTelecoms().add(createTel(ContactType.MC, "112"));
        request.setDataEnterer(createDateEntererElement());

        skrClient.updateContactInformation20200728(request);
    }


    public enum ContactType {
        H, WP, MC
    }

    public void addDentist(String cpr) throws DGWSFault {
        org.hl7.v3.ObjectFactory objectFactory = new org.hl7.v3.ObjectFactory();
        CreateHealthProviderRequest request = new CreateHealthProviderRequest();
        request.setId(createIdElement(cpr));
        HealthProvider healthProvider = new HealthProvider();
        CD providerType = new CD();
        providerType.setCode("tandlæge");
        providerType.setCodeSystem("1.2.208.184.15.12");
        providerType.setDisplayName("dentist");
        healthProvider.setProviderType(providerType);
        POCDMT000040Organization organisation = new POCDMT000040Organization();
        II organisationType = new II();
        organisationType.setAssigningAuthorityName("fds");
        organisationType.setExtension("123456");
        organisationType.setRoot("1.2.208.176.1.1");
        organisation.getIds().add(organisationType);
        ON organisationName = new ON();
        EnFamily family = new EnFamily();
        family.getContent().add("My Dentist");
        organisationName.getContent().add("family");
        organisation.getNames().add(organisationName);
        organisation.getTelecoms().add(createTel());

        AD address = new AD();
        address.setIsNotOrdered(false);
        address.getUses().add("H");
        AdxpStreetAddressLine streetAddress = new AdxpStreetAddressLine();
        streetAddress.getContent().add("Street Name 2");
        AdxpPostalCode postal = new AdxpPostalCode();
        postal.getContent().add("8000");
        AdxpCity city = new AdxpCity();
        city.getContent().add("Aarhus C");
        AdxpCountry country = new AdxpCountry();
        country.getContent().add("Denmark");
        address.getContent().add(objectFactory.createADStreetAddressLine(streetAddress));
        address.getContent().add(objectFactory.createADPostalCode(postal));
        address.getContent().add(objectFactory.createADCity(city));
        address.getContent().add(objectFactory.createADCountry(country));
        organisation.getAddrs().add(address);

        healthProvider.setOrganization(organisation);
        request.setHealthProvider(healthProvider);
        request.setDataEnterer(createDateEntererElement());

        skrClient.createHealthProvider20200728(request);
    }

    public void deleteDentist(String cpr, String dentistIdString) throws DGWSFault {
        if (dentistIdString == null){
            return;
        }
        DeleteHealthProviderRequest request = new DeleteHealthProviderRequest();
        request.setId(createIdElement(cpr));

        II healthProviderId = createId(dentistIdString);

        request.setHealthProviderId(healthProviderId);
        request.setDataEnterer(createDateEntererElement());

        try {
            skrClient.deleteHealthProvider20200728(request);
        } catch (Exception e) {
            if (!e.getMessage().contains("did not match any current data")) {
                throw e;
            }
        }
    }

    private TEL createTel(ContactType contactType, String phone) {
        TEL tel = new TEL();
        tel.setValue("tel:" + phone);
        tel.getUses().add(contactType.toString());

        return tel;
    }

    private TEL createTel() {
        return createTel(ContactType.WP, "911");
    }


    public void addRelative(String cpr) throws DGWSFault {
        CreateRelativesRequest request = new CreateRelativesRequest();
        request.setId(createIdElement(cpr));
        request.setDataEnterer(createDateEntererElement());
        RelatedPerson relatedPerson = new RelatedPerson();
        CD relationshipType = new CD();
        relationshipType.setCode("barn");
        relationshipType.setCodeSystem("1.2.208.184.15.4");
        relatedPerson.setRelationshipType(relationshipType);
        POCDMT000040AssociatedEntity associatedEntity = new POCDMT000040AssociatedEntity();
        associatedEntity.getClassCodes().add("CON");

        associatedEntity.getTelecoms().add(createTel(ContactType.H, "112"));
        associatedEntity.getTelecoms().add(createTel(ContactType.WP, "911"));
        associatedEntity.getTelecoms().add(createTel(ContactType.MC, "118"));
        POCDMT000040Person associatedPerson = new POCDMT000040Person();
        associatedPerson.getNames().add(createGivenAndFamilyName("Relative", "Name"));

        associatedEntity.setAssociatedPerson(associatedPerson);
        relatedPerson.setAssociatedEntity(associatedEntity);
        relatedPerson.setNote("En Note");
        request.getRelatedPersons().add(relatedPerson);

        CreateRelativesResponse response = skrClient.createRelatives20200728(request);
    }

    public void deleteRelative(String cpr, String... relatives) throws DGWSFault {
        DeleteRelativesRequest request = new DeleteRelativesRequest();
        request.setId(createIdElement(cpr));
        request.setDataEnterer(createDateEntererElement());

        Arrays.stream(relatives).forEach(x -> request.getRelativeIds().add(createRelatedId(x)));

        DeleteRelativesResponse response = skrClient.deleteRelatives20200728(request);
    }

    private II createRelatedId(String uuid) {
        II ii = new II();
        ii.setExtension(uuid);
        ii.setRoot("1.2.208.184.15.3");
        ii.setAssigningAuthorityName("FSK");

        return ii;
    }

    public void updateRelative(String cpr) throws DGWSFault {
        UpdateRelativesRequest request = new UpdateRelativesRequest();
        request.setId(createIdElement(cpr));
        request.setDataEnterer(createDateEntererElement());
        RelatedPerson relatedPerson = new RelatedPerson();
        CD relationshipType = new CD();
        relationshipType.setCode("barn");
        relationshipType.setCodeSystem("1.2.208.184.15.4");
        relatedPerson.setRelationshipType(relationshipType);
        POCDMT000040AssociatedEntity associatedEntity = new POCDMT000040AssociatedEntity();
        associatedEntity.getClassCodes().add("CON");

        associatedEntity.getTelecoms().add(createTel());
        POCDMT000040Person associatedPerson = new POCDMT000040Person();
        associatedPerson.getNames().add(createGivenAndFamilyName("Relative", "Name"));

        associatedEntity.setAssociatedPerson(associatedPerson);
        relatedPerson.setAssociatedEntity(associatedEntity);
        relatedPerson.setNote("En Note");
        II id = new II();
        id.setExtension("786769c3-2ec0-4692-92e0-8db5f3c23e99");
        id.setRoot("1.2.208.184.15.3");
        relatedPerson.setId(id);
        request.getRelatedPersons().add(relatedPerson);

        UpdateRelativesResponse response = skrClient.updateRelatives20200728(request);
    }

    public POCDMT000040ClinicalDocument getSkr(String cpr) throws DGWSFault {
        GetPersonalDataCardRequest request = new GetPersonalDataCardRequest();
        request.setId(createIdElement(cpr));

        GetPersonalDataCardResponse response = skrClient.getPersonalDataCard20200728(request);
        return response.getClinicalDocument();
    }

    private PN createGivenAndFamilyName(String relative, String name) {
        PN pn = new PN();

        EnGiven enGiven = new EnGiven();
        enGiven.getContent().add("Casper");
        pn.getContent().add(new org.hl7.v3.ObjectFactory().createENGiven(enGiven));

        EnFamily enFamily = new EnFamily();
        enFamily.getContent().add("Larsen");
        
        pn.getContent().add(new org.hl7.v3.ObjectFactory().createENFamily(enFamily));

        return pn;
    }

    private POCDMT000040Author createDateEntererElement() {
        POCDMT000040Author dataEnterer = new POCDMT000040Author();
        TS ts = new TS();
        ts.setValue(ZonedDateTime.now().format(DateTimeFormatter.ofPattern("yyyyMMddHHmmssZ"))); //
        dataEnterer.setTime(ts);
        POCDMT000040AssignedAuthor assignedAuthor = new POCDMT000040AssignedAuthor();
        assignedAuthor.getIds().add(createIdElement("0309691444"));
        POCDMT000040Person assignedPerson = new POCDMT000040Person();
        PN pn = new PN();

        EnGiven enGiven = new EnGiven();
        enGiven.getContent().add("Sonja");
        pn.getContent().add(new org.hl7.v3.ObjectFactory().createENGiven(enGiven));

        EnFamily enFamily = new EnFamily();
        enFamily.getContent().add("Bach");

        pn.getContent().add(new org.hl7.v3.ObjectFactory().createENFamily(enFamily));

        assignedPerson.getNames().add(pn);
        assignedAuthor.setAssignedPerson(assignedPerson);
        dataEnterer.setAssignedAuthor(assignedAuthor);

        return dataEnterer;
    }

    private II createIdElement(String cpr) {
        II ii = new II();
        ii.setAssigningAuthorityName("CPR");
        ii.setExtension(cpr);
        ii.setRoot("1.2.208.176.1.2");
        return ii;
    }
}
