/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.dgws;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.xml.XmlUtil;
import org.springframework.core.io.Resource;
import org.w3c.dom.Document;

import java.io.IOException;
import java.io.InputStream;
import java.security.*;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Calendar;
import java.util.Date;
import java.util.Properties;

public abstract class DgwsNspSosiIdCardHelper implements SosiIdCardHandler {

    public static final String LAKESIDE_CVR_NUMBER = "25450442";
    public static final String LAKESIDE_A_S = "LAKESIDE A/S";
    public static final String CPR = "2708599967";
    public static final boolean CONSENT_OVERRIDE = false;
    public static final String SSI_CVR_NUMBER = "46837428";
    public static final String SSI = "Statens Serum Institut";

	protected CredentialVaultWrapper credentialVault;

	protected STSRequestHelper requestHelper;

	protected SOSIFactory sosiFactory;

	protected IDCard cachedCard;

	public DgwsNspSosiIdCardHelper(Resource keystore, String password, String alias) throws IOException {
		this.x509 = getX509Wrapper(keystore, password, alias);
		this.credentialVault = new CredentialVaultWrapper(x509.getCertificate(), x509.getPrivateKey());

		Properties props = new Properties(System.getProperties());
		props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, Boolean.toString(true));

		this.requestHelper = new STSRequestHelper();
		this.sosiFactory = new SOSIFactory(new SOSITestFederation(new Properties()), this.credentialVault, props);
	}


	public Document getSosiIdCardRequest() {

		SecurityTokenRequest securityTokenRequest = sosiFactory.createNewSecurityTokenRequest();
		securityTokenRequest.setIDCard(createNewIdCard(sosiFactory));
		Document doc = securityTokenRequest.serialize2DOMDocument();

		if (signSosiCardRequests()) {
			SignatureConfiguration signatureConfiguration = new SignatureConfiguration(new String[] { IDValues.IDCARD }, IDValues.IDCARD, IDValues.id);
			SignatureUtil.sign(SignatureProviderFactory.fromCredentialVault(credentialVault), doc, signatureConfiguration);
		}

		return doc;		
	}

    protected boolean signSosiCardRequests() {
        return true;
    }

	public abstract SystemIDCard createNewIdCard(SOSIFactory sosi);

	public IDCard getToken() throws IOException {

		Calendar cal = Calendar.getInstance();
		cal.add(Calendar.HOUR_OF_DAY, 23); //needs to be syncronized with the idcard.max.age.hours in application.properties. If age.hours set to 1, then hour_of_day needs to be 23. It will then run for aprox 50 min on same idcard.
		Date timeFromNow = cal.getTime();

		if (cachedCard != null && cachedCard.getExpiryDate().after(timeFromNow)) {
			return cachedCard;
		}

		Document requestXml = getSosiIdCardRequest();
		String requestXmlString = XmlUtil.node2String(requestXml, false, true);

		String responseXml = requestHelper.sendRequest(sosiFactory, requestXmlString);
		SecurityTokenResponse securityTokenResponse = sosiFactory.deserializeSecurityTokenResponse(responseXml);

		if (securityTokenResponse.isFault() || securityTokenResponse.getIDCard() == null) {
			throw new RuntimeException("No ID card :-(");
		}
		else {
			SystemIDCard stsSignedIdCard = (SystemIDCard) securityTokenResponse.getIDCard();
			cachedCard = stsSignedIdCard;
			return cachedCard;
		}
	}

	public SOSIFactory getSosiFactory() {
		return this.sosiFactory;
	}

	protected X509Wrapper x509;

	public static X509Wrapper getX509Wrapper(Resource keystoreResource, String password, String alias) throws IOException {
		try {
			InputStream in = keystoreResource.getInputStream();
			KeyStore keyStore = KeyStore.getInstance("JKS");
			keyStore.load(in, password.toCharArray());
			X509Certificate certificate = (X509Certificate) keyStore.getCertificate(alias);
			PrivateKey privateKey = (PrivateKey) keyStore.getKey(alias, password.toCharArray());
			return new X509Wrapper(certificate, privateKey);
		} catch (KeyStoreException | NoSuchAlgorithmException | CertificateException | UnrecoverableKeyException e) {
			throw new IOException("Error loading certificate and private key from " + keystoreResource, e);
		}
	}
	
	@Override
	public Document getSosiDocument() {
		try {
			Request request = getSosiFactory().createNewRequest(false, null);
			request.setIDCard(getToken());
			return request.serialize2DOMDocument();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}
}
